/*
Unnamed DNS Server (http://eder.us/projects/unnamed/
Copyright (C) 2012 David Eder

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <limits.h>
#include <fcntl.h>
#include "unnamed.h"

#ifdef USE_PTHREADS
  static pthread_mutex_t gc_mutex = PTHREAD_MUTEX_INITIALIZER;
#endif

typedef struct
{
  rr_t *head;
  #ifdef USE_PTHREADS
    pthread_mutex_t mutex;
  #endif
} cache_list_t;

#define BUCKETS 1024

typedef struct
{
  cache_list_t buckets[BUCKETS];
} cache_hash_t;

typedef struct
{
  address_t *whitelist_addr;
  unsigned int whitelist_addr_count;
  address_t *forward_addr;
  unsigned int forward_addr_count;
  double forward_timeout;
  uint16_t max_udp_size;
  uint32_t ttl;
  int trace;
} module_data_t;

/* cache is static */
static cache_hash_t cache;
static uint32_t gc_last = 0;
static uint32_t gc_interval = 300;

static rr_t* dup_rr(state_t* state, const rr_t *rr)
{
  rr_t *ret;
  if((ret = (rr_t*)calloc(1, sizeof(rr_t))) == NULL) return NULL;
  ret->rtype = rr->rtype;
  ret->rclass = rr->rclass;
  ret->expires = rr->expires;
  ret->next = NULL;
  if(state->buffer_copy(&ret->name.name, &rr->name.name) != 0)
  {
    state->buffer_free(&ret->name.name);
    free(ret);
    return NULL;
  }
  if(rr->rclass == CLASS_IN)
  {
    switch(rr->rtype)
    {
      case TYPE_A:
      case TYPE_AAAA:
      {
        memcpy(&ret->p.address, &rr->p.address, sizeof(rr->p.address));
      } break;
      case TYPE_CNAME:
      {
        if(state->buffer_copy(&ret->p.cname.name, &rr->p.cname.name) != 0)
        {
          state->buffer_free(&ret->p.cname.name);
          state->buffer_free(&ret->name.name);
          free(ret);
          return NULL;
        }
      } break;
      case TYPE_MX:
      {
        ret->p.mx.preference = rr->p.mx.preference;
        if(state->buffer_copy(&ret->p.mx.exchange.name, &rr->p.mx.exchange.name) != 0)
        {
          state->buffer_free(&ret->p.mx.exchange.name);
          state->buffer_free(&ret->name.name);
          free(ret);
          return NULL;
        }
      } break;
      case TYPE_NS:
      {
        if(state->buffer_copy(&ret->p.ns.name, &rr->p.ns.name) != 0)
        {
          state->buffer_free(&ret->p.ns.name);
          state->buffer_free(&ret->name.name);
          free(ret);
          return NULL;
        }
      } break;
      case TYPE_PTR:
      {
        if(state->buffer_copy(&ret->p.ptr.name, &rr->p.ptr.name) != 0)
        {
          state->buffer_free(&ret->p.ptr.name);
          state->buffer_free(&ret->name.name);
          free(ret);
          return NULL;
        }
      } break;
      case TYPE_SOA:
      {
        if(state->buffer_copy(&ret->p.soa.mname.name, &rr->p.soa.mname.name) != 0)
        {
          state->buffer_free(&ret->p.soa.mname.name);
          state->buffer_free(&ret->name.name);
          free(ret);
          return NULL;
        }
        if(state->buffer_copy(&ret->p.soa.rname.name, &rr->p.soa.rname.name) != 0)
        {
          state->buffer_free(&ret->p.soa.rname.name);
          state->buffer_free(&ret->p.soa.mname.name);
          state->buffer_free(&ret->name.name);
          free(ret);
          return NULL;
        }
        ret->p.soa.serial = rr->p.soa.serial;
        ret->p.soa.refresh = rr->p.soa.refresh;
        ret->p.soa.retry = rr->p.soa.retry;
        ret->p.soa.expire = rr->p.soa.expire;
        ret->p.soa.minimum = rr->p.soa.minimum;
      } break;
      case TYPE_SPF:
      {
        if(state->buffer_copy(&ret->p.spf, &rr->p.spf) != 0)
        {
          state->buffer_free(&ret->p.spf);
          state->buffer_free(&ret->name.name);
          free(ret);
          return NULL;
        }
      } break;
      case TYPE_TXT:
      {
        if(state->buffer_copy(&ret->p.txt, &rr->p.txt) != 0)
        {
          state->buffer_free(&ret->p.txt);
          state->buffer_free(&ret->name.name);
          free(ret);
          return NULL;
        }
      } break;
      case TYPE_SRV:
      {
        ret->p.srv.priority = rr->p.srv.priority;
        ret->p.srv.weight = rr->p.srv.weight;
        ret->p.srv.port = rr->p.srv.port;
        if(state->buffer_copy(&ret->p.srv.target.name, &rr->p.srv.target.name) != 0)
        {
          state->buffer_free(&ret->p.srv.target.name);
          state->buffer_free(&ret->name.name);
          free(ret);
          return NULL;
        }
      } break;
      default: break;
    }
  }
  return ret;
}

static int rr_cmp(state_t *state, const rr_t *a, const rr_t *b)
{
  if(a->rtype != b->rtype || a->rclass != b->rclass) return -1;
  if(state->buffer_cmp(&a->name.name, &b->name.name) != 0) return -1;
  if(a->rclass == CLASS_IN)
  {
    switch(a->rtype)
    {
      case TYPE_A:
      case TYPE_AAAA:
      {
        return memcmp(&a->p.address, &b->p.address, sizeof(a->p.address));
      }
      case TYPE_CNAME:
      {
        return state->buffer_cmp(&a->p.cname.name, &b->p.cname.name);
      }
      case TYPE_MX:
      {
        if(a->p.mx.preference != b->p.mx.preference) return -1;
        return state->buffer_cmp(&a->p.mx.exchange.name, &b->p.mx.exchange.name);
      }
      case TYPE_NS:
      {
        return state->buffer_cmp(&a->p.ns.name, &b->p.ns.name);
      }
      case TYPE_PTR:
      {
        return state->buffer_cmp(&a->p.ptr.name, &b->p.ptr.name);
      }
      case TYPE_SOA:
      {
        if(state->buffer_cmp(&a->p.soa.mname.name, &b->p.soa.mname.name) != 0) return -1;
        if(state->buffer_cmp(&a->p.soa.rname.name, &b->p.soa.rname.name) != 0) return -1;
        if(a->p.soa.serial != b->p.soa.serial || a->p.soa.refresh != b->p.soa.refresh || a->p.soa.retry != b->p.soa.retry || a->p.soa.expire != b->p.soa.expire || a->p.soa.minimum != b->p.soa.minimum) return -1;
      } break;
      case TYPE_SPF:
      {
        return state->buffer_cmp(&a->p.spf, &b->p.spf);
      }
      case TYPE_TXT:
      {
        return state->buffer_cmp(&a->p.txt, &b->p.txt);
      }
      case TYPE_SRV:
      {
        if(a->p.srv.priority != b->p.srv.priority) return -1;
        if(a->p.srv.weight != b->p.srv.weight) return -1;
        if(a->p.srv.port != b->p.srv.port) return -1;
        if(state->buffer_cmp(&a->p.srv.target.name, &b->p.srv.target.name) != 0) return -1;
      }
      default: break;
    }
  }
  return 0;
}

static void collect_garbage(int __attribute__((unused)) module_index, state_t *state)
{
  #ifdef USE_PTHREADS
    (void)pthread_mutex_lock(&gc_mutex);
  #endif
  {
    uint32_t now = (uint32_t)time(NULL);
    if(gc_last + gc_interval < now)
    {
      int i;
      for(i = 0; i < BUCKETS; i++)
      {
        rr_t **node;
        #ifdef USE_PTHREADS
          (void)pthread_mutex_lock(&cache.buckets[i].mutex);
        #endif
        node = &cache.buckets[i].head;
        while(*node != NULL)
        {
          if((*node)->expires <= now)
          {
            rr_t *next = (*node)->next;
            state->free_rr(node);
            *node = next;
          }
          else
          {
            node = &(*node)->next;
          }
        }
        #ifdef USE_PTHREADS
          (void)pthread_mutex_unlock(&cache.buckets[i].mutex);
        #endif
      }
      gc_last = now;
    }
  }
  #ifdef USE_PTHREADS
    (void)pthread_mutex_unlock(&gc_mutex);
  #endif
}

static uint16_t hash_rr(const buffer_t *name)
{
  int i;
  unsigned int h = 0;
  for(i = 0; i < (int)name->length; i++)
  {
    h = (unsigned int)(33 * h + name->value[i]);
  }
  return (uint16_t)(h % BUCKETS);
}

static int cache_rr(state_t *state, int module_index, rr_t *rr)
{
  uint16_t bucket;
  rr_t **node;
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  int i;
  if(rr->rclass != CLASS_IN) return -1;
  for(i = 0; i < (int)rr->name.name.length; i++)  if(rr->name.name.value[i] == (uint8_t)'\\' || rr->name.name.value[i] == (uint8_t)'\'') return -1;
  if(state->is_name_black(state, &rr->name.name) != 0) return -1;
  DEBUG_IF(module->trace) (void)state->lprintf(TRACE, "%s", state->type_name(rr->rtype));
  switch(rr->rtype)
  {
    case TYPE_A:
    case TYPE_AAAA:
    {
      char buffer[IPCHAR_SIZE];
      buffer[0] = '\0';
      DEBUG_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => %s", (char*)rr->name.name.value, state->class_name(rr->rclass), state->type_name(rr->rtype), state->addr2char(buffer, &rr->p.address));
      if(state->is_ip_black(state, &rr->p.address) != 0) rr->status = RCODE_NAME_ERROR;
    } break;
    case TYPE_CNAME:
    {
      DEBUG_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => %s", (char*)rr->name.name.value, state->class_name(rr->rclass), state->type_name(rr->rtype), (char*)rr->p.cname.name.value);
      if(state->is_name_black(state, &rr->p.cname.name) != 0) rr->status = RCODE_NAME_ERROR;
    } break;
    case TYPE_MX:
    {
      DEBUG_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => %u %s", (char*)rr->name.name.value, state->class_name(rr->rclass), state->type_name(rr->rtype), rr->p.mx.preference, (char*)rr->p.mx.exchange.name.value);
      if(state->is_name_black(state, &rr->p.mx.exchange.name) != 0) rr->status = RCODE_NAME_ERROR;
    } break;
    case TYPE_NS:
    {
      DEBUG_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => %s", (char*)rr->name.name.value, state->class_name(rr->rclass), state->type_name(rr->rtype), (char*)rr->p.ns.name.value);
      if(state->is_name_black(state, &rr->p.ns.name) != 0) rr->status = RCODE_NAME_ERROR;
    } break;
    case TYPE_PTR:
    {
      DEBUG_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => %s", (char*)rr->name.name.value, state->class_name(rr->rclass), state->type_name(rr->rtype), (char*)rr->p.ptr.name.value);
      if(state->is_name_black(state, &rr->p.ptr.name) != 0) rr->status = RCODE_NAME_ERROR;
    } break;
    case TYPE_SOA:
    {
      DEBUG_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => mname=%s rname=%s", (char*)rr->name.name.value, state->class_name(rr->rclass), state->type_name(rr->rtype), (char*)rr->p.soa.mname.name.value, (char*)rr->p.soa.rname.name.value);
      if(state->is_name_black(state, &rr->p.soa.mname.name) != 0 || state->is_name_black(state, &rr->p.soa.rname.name) != 0) rr->status = RCODE_NAME_ERROR;
    } break;
    case TYPE_SPF:
    {
      DEBUG_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => (%u) %s", (char*)rr->name.name.value, state->class_name(rr->rclass), state->type_name(rr->rtype), rr->p.spf.length, (char*)rr->p.spf.value);
    } break;
    case TYPE_TXT:
    {
      DEBUG_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => (%u) %s", (char*)rr->name.name.value, state->class_name(rr->rclass), state->type_name(rr->rtype), rr->p.txt.length, (char*)rr->p.txt.value);
    } break;
    case TYPE_SRV:
    {
      DEBUG_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s => priority=%u %u %u %s", (char*)rr->name.name.value, state->class_name(rr->rclass), state->type_name(rr->rtype), rr->p.srv.priority, rr->p.srv.weight, rr->p.srv.port, (char*)rr->p.srv.target.name.value);
      if(state->is_name_black(state, &rr->p.srv.target.name) != 0) rr->status = RCODE_NAME_ERROR;
    } break;
    default:
    {
      return -1;
    }
  }

  bucket = hash_rr(&rr->name.name);

  #ifdef USE_PTHREADS
    (void)pthread_mutex_lock(&cache.buckets[bucket].mutex);
  #endif
  node = &cache.buckets[bucket].head;
  while(*node != NULL)
  {
    if(rr_cmp(state, *node, rr) == 0)
    {
      (*node)->expires = rr->expires;
      state->free_rr(&rr);
      rr = NULL;
      break;
    }
    node = &(*node)->next;
  }
  if(rr != NULL)
  {
    rr->next = cache.buckets[bucket].head;
    cache.buckets[bucket].head = rr;
  }
  #ifdef USE_PTHREADS
    (void)pthread_mutex_unlock(&cache.buckets[bucket].mutex);
  #endif
  return 0;
}

static answer_t answer_cache(int module_index, client_t *client, rr_section_t rr_section, const buffer_t *name, uint16_t type, uint16_t dclass, int level)
{
  answer_t ret = ANSWER_UNSATISFIED;
  int not_found = 0, name_error = 0, dups = 0, i;
  rr_t *rr, *duprr[MAX_RRS];
  uint16_t bucket;
  state_t *state = client->state;
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  uint32_t now = (uint32_t)time(NULL);
  bucket = hash_rr(name);
  duprr[0] = NULL;

  if(client->header.flags.ra == 0) return ANSWER_UNSATISFIED;

  #ifdef USE_PTHREADS
    (void)pthread_mutex_lock(&cache.buckets[bucket].mutex);
  #endif

  for(rr = cache.buckets[bucket].head; rr != NULL && dups < MAX_RRS; rr = rr->next)
  {
    if(rr->rtype == type && rr->rclass == dclass && rr->expires > now && state->buffer_cmp(&rr->name.name, name) == 0)
    {
      DEBUG_IF(module->trace) (void)state->lprintf(TRACE, "cache %s %s %s => %d %s", (char*)rr->name.name.value, state->class_name(rr->rclass), state->type_name(rr->rtype), rr->status, state->rcode_name(rr->status));
      if(rr->status == RCODE_NO_ERROR)
      {
        if((duprr[dups] = dup_rr(state, rr)) != NULL) dups++;
      }
      else if(rr->status == RCODE_NAME_ERROR)
      {
        name_error = 1;
      }
      else if(rr->status == RCODE_NOT_FOUND)
      {
        not_found = 1;
      }
    }
  }

  #ifdef USE_PTHREADS
    (void)pthread_mutex_unlock(&cache.buckets[bucket].mutex);
  #endif

  for(i = 0; i < dups; i++)
  {
    if(duprr[i]->rclass == CLASS_IN && dclass == CLASS_IN && duprr[i]->rtype == TYPE_CNAME && type == TYPE_CNAME && level > 0)
    {
      buffer_t cname;
      buffer_init(&cname);
      if(state->buffer_copy(&cname, &duprr[i]->p.cname.name) == 0)
      {
        if(state->add_rr(client, rr_section, &duprr[i]) == 0) ret = ANSWER_SATISFIED;
        (void)answer_cache(module_index, client, rr_section, &cname, TYPE_CNAME, CLASS_IN, level - 1);
        (void)answer_cache(module_index, client, rr_section, &cname, TYPE_A, CLASS_IN, level - 1);
        (void)answer_cache(module_index, client, rr_section, &cname, TYPE_AAAA, CLASS_IN, level - 1);
        (void)answer_cache(module_index, client, rr_section, &cname, TYPE_PTR, CLASS_IN, level - 1);
        (void)answer_cache(module_index, client, rr_section, &cname, TYPE_SOA, CLASS_IN, level - 1);
      }
      state->buffer_free(&cname);
    }
    else
    {
      if(state->add_rr(client, rr_section, &duprr[i]) == 0) ret = ANSWER_SATISFIED;
    }
    state->free_rr(&duprr[i]);
  }

  if(ret == ANSWER_UNSATISFIED)
  {
    if(not_found != 0) ret = ANSWER_NOTFOUND;
    if(name_error != 0) ret = ANSWER_NAMEERROR;
  }

  DEBUG_IF(module->trace) (void)state->lprintf(TRACE, "%s %s %s %d => %s", (char*)name->value, state->class_name(dclass), state->type_name(type), level, state->answer_type_name(ret));

  return ret;
}

static void term(int module_index, state_t *state)
{
  int i;
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  for(i = 0; i < BUCKETS; i++)
  {
    rr_t **node = &cache.buckets[i].head;
    while(*node != NULL)
    {
      rr_t *next = (*node)->next;
      state->free_rr(node);
      *node = next;
    }
  }

  if(module->whitelist_addr != NULL)
  {
    free(module->whitelist_addr);
    module->whitelist_addr_count = 0;
  }
  if(module->forward_addr != NULL)
  {
    free(module->forward_addr);
    module->forward_addr_count = 0;
  }

  free(module);
  state->module_data[module_index] = NULL;
}

static int init(int module_index, state_t *state)
{
  int i;
  buffer_t buffer;
  module_data_t *module = (module_data_t*)calloc(1, sizeof(*module));
  if(module == NULL) return -1;
  state->module_data[module_index] = module;

  module->trace = state->config_get_int(state, "forward.trace", state->trace);
  module->ttl = (uint32_t)state->config_get_int(state, "forward.ttl", DEFAULT_TTL);
  gc_interval = (uint32_t)state->config_get_int(state, "forward.gc_interval", (int)gc_interval);
  module->forward_timeout = state->config_get_double(state, "forward.timeout", 2.0);
  module->max_udp_size = (uint16_t)state->config_get_int(state, "forward.max_udp_size", 512);

  buffer_init(&buffer);
  for(i = 1; ; i++)
  {
    address_t *temp = module->whitelist_addr;
    module->whitelist_addr = (address_t*)realloc(module->whitelist_addr, sizeof(*module->whitelist_addr) * (module->whitelist_addr_count + 1));
    if(module->whitelist_addr == NULL)
    {
      module->whitelist_addr = temp;
      ERROR_IF(module->trace) (void)state->lprintf(TRACE, "realloc failed");
      term(module_index, state);
      return -1;
    }
    if(state->buffer_set(&buffer, "forward.whitelisted.%d.ip", i) != 0)
    {
      ERROR_IF(module->trace) (void)state->lprintf(TRACE, "buffer_set failed");
      term(module_index, state);
      return -1;
    }
    if(state->config_get_ip(state, (char*)buffer.value, &module->whitelist_addr[module->whitelist_addr_count]) != 0) break;
    if(state->buffer_set(&buffer, "forward.whitelisted.%d.port", i) != 0)
    {
      ERROR_IF(module->trace) (void)state->lprintf(TRACE, "buffer_set failed");
      term(module_index, state);
      return -1;
    }
    module->whitelist_addr[module->whitelist_addr_count].a.sockaddr_in.sin_port = htons((uint16_t)state->config_get_int(state, (char*)buffer.value, 53));
    module->whitelist_addr_count++;
  }

  for(i = 1; ; i++)
  {
    address_t *temp = module->forward_addr;
    module->forward_addr = (address_t*)realloc(module->forward_addr, sizeof(*module->forward_addr) * (module->forward_addr_count + 1));
    if(module->forward_addr == NULL)
    {
      module->forward_addr = temp;
      ERROR_IF(module->trace) (void)state->lprintf(TRACE, "realloc failed");
      term(module_index, state);
      return -1;
    }
    if(state->buffer_set(&buffer, "forward.nonwhitelisted.%d.ip", i) != 0)
    {
      ERROR_IF(module->trace) (void)state->lprintf(TRACE, "buffer_set failed");
      term(module_index, state);
      return -1;
    }
    if(state->config_get_ip(state, (char*)buffer.value, &module->forward_addr[module->forward_addr_count]) != 0) break;
    if(state->buffer_set(&buffer, "forward.nonwhitelisted.%d.port", i) != 0)
    {
      ERROR_IF(module->trace) (void)state->lprintf(TRACE, "buffer_set failed");
      term(module_index, state);
      return -1;
    }
    module->forward_addr[module->forward_addr_count].a.sockaddr_in.sin_port = htons((uint16_t)state->config_get_int(state, (char*)buffer.value, 53));
    module->forward_addr_count++;
  }

  for(i = 0; i < BUCKETS; i++)
  {
    #ifdef USE_PTHREADS
      (void)pthread_mutex_init(&cache.buckets[i].mutex, NULL);
    #endif
    cache.buckets[i].head = NULL;
  }
  state->config_free_node(state, "forward");
  return 0;
}

static answer_t answer(int module_index, client_t *client, rr_section_t rr_section, const buffer_t *name, uint16_t type, uint16_t dclass, int lazy)
{
  fd_set rfds;
  struct timeval tv;
  address_t *addr, reply;
  socklen_t reply_len = (socklen_t)sizeof(reply);
  client_t local_client;
  uint16_t id = 0;
  double start;
  answer_t rc;
  state_t *state = client->state;
  module_data_t *module = (module_data_t*)state->module_data[module_index];
  int found = 0, val, i, j, ns, bytes = 0;
  SOCKET sock[16], sock_max = 0;
  int sock_count = 0;

  start = state->microtime();

  #ifdef S_SPLINT_S
    memset(sock, 0, sizeof(sock));
  #endif

  if(client->header.flags.ra == 0) return ANSWER_UNSATISFIED;
  if(state->is_subdomain("unknown", (char*)name->value) != 0) return ANSWER_UNSATISFIED;
  if(state->is_name_black(state, name) != 0) return ANSWER_BLOCKED;

  rc = answer_cache(module_index, client, rr_section, name, type, dclass, 5);
  if(rc == ANSWER_SATISFIED) return ANSWER_SATISFIED;
  if(rc == ANSWER_NAMEERROR) return ANSWER_UNSATISFIED;
  if(rc == ANSWER_NOTFOUND && dclass == CLASS_IN)
  {
    if(type == TYPE_A || type == TYPE_AAAA || type == TYPE_PTR)
    {
      rc = answer_cache(module_index, client, rr_section, name, TYPE_CNAME, dclass, 5);
    }
    (void)answer_cache(module_index, client, rr_section, name, TYPE_SOA, dclass, 5);
    if(rc != ANSWER_SATISFIED) rc = ANSWER_UNSATISFIED;
    return rc;
  }
  if(lazy != 0) return ANSWER_UNSATISFIED;

  addr = module->forward_addr;
  sock_count = (int)module->forward_addr_count;
  if(dclass == CLASS_IN)
  {
    if(type == TYPE_MX || type == TYPE_NS || type == TYPE_SOA || type == TYPE_PTR || type == TYPE_SPF || type == TYPE_TXT || type == TYPE_SRV)
    {
      addr = module->whitelist_addr;
      sock_count = (int)module->whitelist_addr_count;
    }
  }
  if(addr != module->whitelist_addr && state->is_name_white(state, name) != 0)
  {
    addr = module->whitelist_addr;
    sock_count = (int)module->whitelist_addr_count;
  }
  if(sock_count > (int)(sizeof(sock) / sizeof(sock[0]))) sock_count = (int)(sizeof(sock) / sizeof(sock[0]));

  memset(&local_client, 0, sizeof(local_client));
  if(state->init_client(state, &local_client, module->max_udp_size) != 0)
  {
    (void)state->free_client(&local_client, 0);
    ERROR_IF(module->trace) (void)state->lprintf(TRACE, "init_client failed");
    return ANSWER_UNSATISFIED;
  }

  id = state->random_id(state);
  local_client.header.id = id;
  local_client.header.flags.rd = 1;
  local_client.header.qdcount = 0;
  local_client.header.ancount = 0;
  local_client.header.nscount = 0;
  local_client.header.arcount = 0;

  if((local_client.questions[local_client.header.qdcount++] = state->question_create(name, type, dclass)) == NULL)
  {
    ERROR_IF(module->trace) (void)state->lprintf(TRACE, "question_create failed");
    (void)state->free_client(&local_client, 0);
    return ANSWER_UNSATISFIED;
  }

#if 0
  if(module->max_udp_size > 512)
  {
    rr_t *rr;
    WARNING_IF(state->trace) (void)state->lprintf(TRACE, "max_udp_size = %d", module->max_udp_size);
    rr = state->rr_create(NULL, TYPE_OPT, module->max_udp_size, 0, RCODE_NO_ERROR);
    if(rr != NULL)
    {
      (void)state->add_rr(&local_client, RR_AR, &rr);
      state->free_rr(&rr);
    }
  }
#endif

  if(state->write_packet(&local_client) != 0)
  {
    ERROR_IF(module->trace) (void)state->lprintf(TRACE, "write_packet failed");
    (void)state->free_client(&local_client, 0);
    return ANSWER_UNSATISFIED;
  }

  val = 1;
  (void)setsockopt(client->socket, SOL_SOCKET, SO_REUSEADDR, (const char*)&val, (socklen_t)sizeof(val));

  FD_ZERO(&rfds);
  for(i = 0; i < sock_count; i++)
  {
    if((sock[i] = socket((int)addr[i].a.sockaddr.sa_family, SOCK_DGRAM, 0)) == INVALID_SOCKET)
    {
      ERROR_IF(module->trace) (void)state->lprintf(TRACE, "socket failed: %s", ERRNO_STRING);
      for(j = 0; j < i; j++) (void)closesocket(sock[j]);
      (void)state->free_client(&local_client, 0);
      return ANSWER_UNSATISFIED;
    }

    if((int)sendto(sock[i], (char*)(local_client.packet.value + PACKET_LENGTH_SIZE), (buf_len)local_client.packet.length - PACKET_LENGTH_SIZE, 0, &addr[i].a.sockaddr, (socklen_t)sizeof(addr[0].a.sockaddr)) != (int)local_client.packet.length - PACKET_LENGTH_SIZE)
    {
      char ip[IPCHAR_SIZE];
      ip[0] = '\0';
      ERROR_IF(module->trace) (void)state->lprintf(TRACE, "sendto %s:%d failed: %s  sock[%d]=%d length=%d", state->addr2char(ip, &addr[i]), ntohs(addr[i].a.sockaddr_in.sin_port), ERRNO_STRING, i, (int)sock[i], (int)local_client.packet.length - PACKET_LENGTH_SIZE);

      /* We want to close all of the sockets we have open.  Thus it should be j <= i.
         Unfortunately gcc doesn't like it.  So we do j < i, followed by another close. */
      for(j = 0; j < i; j++) (void)closesocket(sock[j]);
      (void)closesocket(sock[j]);

      (void)state->free_client(&local_client, 0);
      return ANSWER_UNSATISFIED;
    }
    FD_SET((unsigned int)sock[i], &rfds);
    if(sock[i] > sock_max) sock_max = sock[i];
  }

  (void)state->buffer_resize(&local_client.packet, 512);

  #ifdef S_SPLINT_S
    memset(&tv, 0, sizeof(tv));
  #endif
  tv.tv_sec = (long)module->forward_timeout;
  tv.tv_usec = (long)(1000000.0 * (module->forward_timeout - (double)tv.tv_sec));

  ns = select((int)sock_max+1, &rfds, NULL, NULL, &tv);
  if(ns == -1)
  {
    ERROR_IF(module->trace) (void)state->lprintf(TRACE, "select failed: %s", ERRNO_STRING);
    for(j = 0; j < sock_count; j++) (void)closesocket(sock[j]);
    (void)state->free_client(&local_client, 0);
    return ANSWER_UNSATISFIED;
  }
  if(ns == 0) /* timeout */
  {
    ERROR_IF(module->trace) (void)state->lprintf(TRACE, "select timeout");
    for(j = 0; j < sock_count; j++) (void)closesocket(sock[j]);
    (void)state->free_client(&local_client, 0);
    return ANSWER_UNSATISFIED;
  }

  memset(&reply, 0, sizeof(reply));

  for(i = 0; i < sock_count && ns > 0; i++)
  {
    if(FD_ISSET(sock[i], &rfds))
    {
      ns--;
      if((bytes = (int)recvfrom(sock[i], (char*)local_client.packet.value + PACKET_LENGTH_SIZE, (buf_len)(local_client.packet.allocated - PACKET_LENGTH_SIZE), 0, (struct sockaddr*)&reply, &reply_len)) == -1)
      {
        ERROR_IF(module->trace) (void)state->lprintf(TRACE, "recvfrom failed: %s", ERRNO_STRING);
        continue;
      }
      if(state->addr_cmp(&reply, &addr[i]) != 0)
      {
        char ip[IPCHAR_SIZE];
        ip[0] = '\0';
        WARNING_IF(module->trace) (void)state->lprintf(TRACE, "Got a spoofed packet from %s", state->addr2char(ip, &reply));
        continue;
      }
      break;
    }
  }
  INFO_IF(module->trace) (void)state->lprintf(TRACE, "request took %f seconds for %s %s %s", state->microtime() - start, (char*)name->value, state->class_name(dclass), state->type_name(type));
  for(j = 0; j < sock_count; j++) (void)closesocket(sock[j]);

  if(bytes == 0)
  {
    WARNING_IF(module->trace) (void)state->lprintf(TRACE, "Unable to parse response");
    (void)state->free_client(&local_client, 0);
    return ANSWER_UNSATISFIED;
  }

  local_client.packet.length = (uint32_t)bytes + PACKET_LENGTH_SIZE;

  if(state->parse_packet(&local_client) != 0)
  {
    WARNING_IF(module->trace) (void)state->lprintf(TRACE, "Unable to parse response");
    (void)state->free_client(&local_client, 0);
    return ANSWER_UNSATISFIED;
  }

  if(id != local_client.header.id)
  {
    WARNING_IF(module->trace) (void)state->lprintf(TRACE, "ID mismatch %u != %u", (unsigned int)local_client.header.id, (unsigned int)id);
    (void)state->free_client(&local_client, 0);
    return ANSWER_UNSATISFIED;
  }

  for(i = (int)local_client.header.ancount - 1; i >= 0; i--)
  {
    if(local_client.answers[i]->rtype == type && local_client.answers[i]->rclass == dclass && state->buffer_cmp(&local_client.answers[i]->name.name, name) == 0) found = 1;
    if(cache_rr(state, module_index, local_client.answers[i]) == 0) local_client.answers[i] = NULL;
  }
  for(i = (int)local_client.header.nscount - 1; i >= 0; i--)
  {
    if(cache_rr(state, module_index, local_client.name_servers[i]) == 0) local_client.name_servers[i] = NULL;
  }
  for(i = (int)local_client.header.arcount - 1; i >= 0; i--)
  {
    if(cache_rr(state, module_index, local_client.additional_records[i]) == 0) local_client.additional_records[i] = NULL;
  }

  if(local_client.header.flags.rcode == RCODE_NAME_ERROR)
  {
    rr_t *rr = state->rr_create(name, type, dclass, (uint32_t)time(NULL) + module->ttl, RCODE_NAME_ERROR);
    if(cache_rr(state, module_index, rr) != 0) state->free_rr(&rr);
  }
  else if(found == 0)
  {
    rr_t *rr = state->rr_create(name, type, dclass, (uint32_t)time(NULL) + module->ttl, RCODE_NOT_FOUND);
    if(cache_rr(state, module_index, rr) != 0) state->free_rr(&rr);
  }

  (void)state->free_client(&local_client, 0);

  rc = answer_cache(module_index, client, rr_section, name, type, dclass, 5);

  if(rc == ANSWER_NAMEERROR) return ANSWER_UNSATISFIED;
  if(rc == ANSWER_NOTFOUND && dclass == CLASS_IN)
  {
    if(type == TYPE_A || type == TYPE_AAAA || type == TYPE_PTR)
    {
      rc = answer_cache(module_index, client, rr_section, name, TYPE_CNAME, dclass, 5);
      if(rc != ANSWER_SATISFIED) rc = ANSWER_UNSATISFIED;
    }
    (void)answer_cache(module_index, client, rr_section, name, TYPE_SOA, dclass, 5);
  }

  return rc;
}

module_t unnamed_module = { "forward", NULL, init, answer, collect_garbage, term };
